#!/usr/bin/perl

#---------------------------------------------------------------------
# Copyright (C) 2003 Chad Redman <chad_eu2@katica.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#---------------------------------------------------------------------


# File:     EU2/Parse/EventParser.pm
# Author:   Chad Redman <chad_eu2@katica.org>

# $Id: EventParser.pm,v 1.2 2003/07/22 04:45:04 chad Exp $

package EU2::Parse::EventParser;
use EU2::Parse::GenericParser;

@ISA= ('EU2::Parse::GenericParser');

%db_countries = (ADE=>1, AFG=>1, AKK=>1, ALB=>1, ALD=>1, AMI=>1, ANN=>1, APA=>1, AQU=>1, ARA=>1, ARG=>1, ARK=>1, ARM=>1, ASH=>1, ASS=>1,
        AST=>1, ATH=>1, ATJ=>1, AUV=>1, AYU=>1, AZT=>1, BAD=>1, BAL=>1, BAN=>1, BAY=>1, BEI=>1, BEN=>1, BGL=>1, BOH=>1, BOS=>1,
        BOU=>1, BRA=>1, BRE=>1, BRI=>1, BRZ=>1, BUL=>1, BUR=>1, BYZ=>1, CAL=>1, CAM=>1, CAN=>1, CAS=>1, CAT=>1, CHA=>1, CHE=>1,
        CHG=>1, CHI=>1, CHM=>1, COL=>1, COR=>1, CRE=>1, CRI=>1, CRO=>1, CYP=>1, CYR=>1, DAH=>1, DAI=>1, DAK=>1, DAN=>1, DEL=>1,
        DLH=>1, DUL=>1, EIR=>1, ENG=>1, ETH=>1, FEZ=>1, FIN=>1, FLA=>1, FPR=>1, FRA=>1, FRI=>1, FUC=>1, GEL=>1, GEN=>1, GEO=>1,
        GRA=>1, GRE=>1, GUJ=>1, HAB=>1, HAI=>1, HAN=>1, HAU=>1, HEI=>1, HEL=>1, HES=>1, HOL=>1, HSA=>1, HUN=>1, HUR=>1, HYD=>1,
        INC=>1, IRA=>1, ITA=>1, JAI=>1, JER=>1, KAL=>1, KAR=>1, KAZ=>1, KHM=>1, KLE=>1, KNI=>1, KOL=>1, KON=>1, KOR=>1, KSH=>1,
        KUR=>1, KZK=>1, LAP=>1, LAT=>1, LIT=>1, LOR=>1, LUA=>1, LUX=>1, MAA=>1, MAG=>1, MAH=>1, MAI=>1, MAL=>1, MAM=>1, MAN=>1,
        MCH=>1, MEC=>1, MER=>1, MEX=>1, MIN=>1, MKS=>1, MLC=>1, MLO=>1, MLW=>1, MOG=>1, MOL=>1, MOR=>1, MOS=>1, MTR=>1, MUN=>1,
        MUS=>1, MYA=>1, MYS=>1, NAP=>1, NAT=>1, NAV=>1, NIP=>1, NOG=>1, NOR=>1, NVG=>1, OHI=>1, OLD=>1, OMA=>1, ORI=>1, ORL=>1,
        PAP=>1, PAR=>1, PEG=>1, PER=>1, PFA=>1, PIR=>1, PIS=>1, POL=>1, POM=>1, POR=>1, PRM=>1, PRO=>1, PRU=>1, PSK=>1, PUR=>1,
        QAR=>1, QUE=>1, RAG=>1, REB=>1, ROY=>1, RUS=>1, RYA=>1, SAC=>1, SAR=>1, SAV=>1, SCO=>1, SER=>1, SHA=>1, SHL=>1, SIB=>1,
        SIC=>1, SIE=>1, SLZ=>1, SON=>1, SPA=>1, SPR=>1, STE=>1, STR=>1, SUD=>1, SUZ=>1, SWE=>1, TAU=>1, TEK=>1, THU=>1, TIB=>1,
        TIM=>1, TOS=>1, TRE=>1, TRI=>1, TUN=>1, TUR=>1, TVE=>1, U00=>1, U01=>1, U02=>1, U03=>1, U04=>1, U05=>1, U06=>1, U07=>1,
        U08=>1, U09=>1, U10=>1, U11=>1, U12=>1, U13=>1, U14=>1, U15=>1, U16=>1, U17=>1, U18=>1, U19=>1, UKR=>1, USA=>1, UZB=>1,
        VEN=>1, VIE=>1, VIJ=>1, WAL=>1, WLS=>1, WUR=>1, XHO=>1, ZAN=>1, ZAP=>1, ZIM=>1, ZUL=>1,);


%db_cultures = (abenaki=>1, aborigin=>1, afghani=>1, aka=>1, albanian=>1, aleutian=>1, altai=>1, amazonian=>1, andean=>1,
    anglosaxon=>1, arabic=>1, armenian=>1, ashanti=>1, aztek=>1, baltic=>1, baluchi=>1, bantu=>1, basque=>1,
    bengali=>1, berber=>1, burmanese=>1, canary=>1, cantonese=>1, caribbean=>1, cherokee=>1, cree=>1, creek=>1,
    czech=>1, dakota=>1, delaware=>1, dravidian=>1, dutch=>1, dyola=>1, ethiopian=>1, filippine=>1, french=>1,
    gaelic=>1, georgian=>1, german=>1, greek=>1, guajiro=>1, gujarati=>1, han=>1, hawaiian=>1, hindi=>1, huron=>1,
    iberian=>1, indonesian=>1, inuit=>1, iroquis=>1, italian=>1, japanese=>1, javan=>1, khazak=>1, khmer=>1,
    kongolese=>1, korean=>1, kurdish=>1, laotian=>1, lithuanian=>1, madagasque=>1, magyar=>1, malay=>1, mali=>1,
    maltese=>1, manchu=>1, marathi=>1, mataco=>1, mayan=>1, melanese=>1, mesoamerican=>1, mississippian=>1,
    mongol=>1, naskapi=>1, navajo=>1, none=>1, nubian=>1, patagonian=>1, persian=>1, polish=>1, polynese=>1,
    romanian=>1, russian=>1, ruthenian=>1, scandinavian=>1, senegambian=>1, shawnee=>1, shona=>1, sikh=>1,
    slavonic=>1, slovak=>1, somali=>1, swahili=>1, swiss=>1, syrian=>1, teremembe=>1, thai=>1, tibetan=>1,
    tuareg=>1, tupinamba=>1, turkish=>1, ugric=>1, uzbehk=>1, vietnamese=>1, yorumba=>1, zapotek=>1,);

%db_religions = (catholic=>1, counterreform=>1, protestant=>1, reformed=>1, orthodox=>1, sunni=>1, shiite=>1,
    confucianism=>1, buddhism=>1, hinduism=>1, pagan=>1);

%db_continents = (europe=>1, america=>1, asia=>1, africa=>1, oceania=>1,);

%db_techs = (latin=>1, torthodox=>1, muslim=>1, china=>1, exotic=>1,);

%db_diplo =  (aristocracy=>1, centralization=>1, innovative=>1, mercantilism=>1, land=>1, offensive=>1, quality=>1, serfdom=>1, );

%db_buildings = (shipyard=>1, barrack=>1, bailiff=>1, courthouse=>1, cityrights=>1, );

## Note: -1 means a random manufactury
%db_manufacturies = ( navalequipment=>1, luxury=>1, goods=>1, refinery=>1, weapons=>1, '-1'=>1, );



### Dates use text for month. This maps each to a number
%db_months = qw(
    january    01
    february   02
    march      03
    april      04
    may        05
    june       06
    july       07
    august     08
    september  09
    october    10
    november   11
    december   12
);

sub new {
    my ($pkg) = @_;
    my $self = EU2::Parse::GenericParser::new();
    $self->{EVENTS} = [];

    return bless $self, $pkg;
}

sub events {  my $self = shift; return $self->{EVENTS}; }
sub addEvent {  my ($self, $obj) = @_; push(@{$self->{EVENTS}}, $obj) }

sub matches      { return EU2::Parse::GenericParser::matches('EU2::Parse::EventParser', @_); }
sub warnBadState { return EU2::Parse::GenericParser::warnBadState(@_); }
sub balanceParen { return EU2::Parse::GenericParser::balanceParen(@_); }
sub optCheckRange { return EU2::Parse::GenericParser::optCheckRange(@_); }
#sub _abortParseError { return EU2::Parse::GenericParser::_abortParseError(@_); }


sub parse {
    my $self = shift;
    $self->_tokenize;

    my $toks = $self->{symbols};

    while (@$toks) {
        my $event = Event->new;
        if ($event->parse($toks)) {
            $self->addEvent($event);
        } else {
            last;
        #    Log(0, "DEBUG? Warning: Expected 'event = {' at line " . $toks->[0]->line . " char " . $toks->[0]->char);
        #    warnBadState(ref $self, '-TOP-', $toks);
        #    return _abortParseError($toks);
        }
    }
}


sub tokenMatches {
    my ($self, $match, $src) = @_;

    if ($match =~ /^:ctag(,-1)?:$/) {
        next if (($1 eq ',-1') and ($src == -1 or $src == -3));
        return 0 if $src !~ /^[A-Za-z0-9]{3}$/;
        if (defined $db_countries{uc $src}) {
            return 1;
        } else {
            $self->Log("Country tag '$src' invalid");
            return 0;
        }
    } elsif ($match =~ /^:culture(,-1)?:$/) {
        next if (($1 eq ',-1') and ($src == -1));
        if (defined $db_cultures{lc $src}) {
            return 1;
        } else {
            $self->Log("Culture '$src' invalid");
            return 0;
        }
    } elsif ($match eq ':religion:') {
        if (defined $db_religions{lc $src}) {
            return 1;
        } else {
            $self->Log("Religion '$src' invalid");
            return 0;
        }
    } elsif ($match eq ':continent:') {
        if (defined $db_continents{lc $src}) {
            return 1;
        } else {
            $self->Log("Continent '$src' invalid");
            return 0;
        }
    } elsif ($match eq ':tech:') {
        if (defined $db_techs{lc $src}) {
            return 1;
        } else {
            $self->Log("Tech group '$src' invalid");
            return 0;
        }
    } elsif ($match eq ':diplo:') {
        if (defined $db_diplo{lc $src}) {
            return 1;
        } else {
            $self->Log("Diplomacy category '$src' invalid");
            return 0;
        }
    } elsif ($match eq ':building:') {
        if (defined $db_buildings{lc $src}) {
            return 1;
        } else {
            $self->Log("Building type '$src' invalid");
            return 0;
        }
    } elsif ($match eq ':manufactury:') {
        if (defined $db_manufacturies{lc $src}) {
            return 1;
        } else {
            $self->Log("Manufactury type '$src' invalid");
            return 0;
        }

    }

    -1;    # Indicates caller can continue its own attempts to match
}



#-------------------------------------
package Event;

sub Log {
    my $obj = shift;
    EU2::Parse::EventParser->Log(@_);
}

sub new {
    my $pkg = shift;
    my $self = {
        'LINE' => undef,    # Line no. on which the block starts
        'ID' => undef,
        'NAME' => undef,
        'DESC' => undef,
        'MONARCH_REFS' => [],
        'LEADER_REFS' => [],
        'EVENT_REFS' => [],
        'ACTIONNAME_REFS' => [],
    };

    return bless $self, $pkg;
}


sub parse {
    my ($obj, $toks) = @_;
    my $state = '<start>';

    my $startline = $toks->[0]->line;
    $obj->{LINE} = $startline;
    my @tmpdate;  ## Store the YMD of dates while parsing state <date>

    while ($state ne '<end>') {
        if (scalar @$toks == 0) {  ## nothing left in the whole file
            $state = '<end>';
            next;
        }

        if ($state eq '<start>') {
            if ( EU2::Parse::EventParser::matches( $toks, qw(event = { ))) {
                $state = '<event>';
                splice (@$toks, 0, 3);
            } else {
                ## Didn't event match 'event='. Probably not a valid event file.
                ## Clear it all out
                EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                while (@$toks) {
                    if ( EU2::Parse::EventParser::matches( $toks, qw(event = { ))) {
                        last;
                    }
                    shift @$toks;
                }
                if (!@$toks) {
                    $obj->Log("No event found before end of file");
                    #$state = '<end>';
                    return 0;
                }
            }
        }

        elsif ($state eq '<event>') {
            # id
            if ( EU2::Parse::EventParser::matches( $toks, qw( id = :int,0,4294967296: ))) {
                $obj->{ID} = 1*$toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            # random
            if ( EU2::Parse::EventParser::matches( $toks, qw( random = (yes|no) ))) {
                splice (@$toks, 0, 3);
                next;
            }

            # country tag
            if ( EU2::Parse::EventParser::matches( $toks, qw( country = :ctag: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            # province tag. Reported to only work since 1.06. '-1' is invalid
            if ( EU2::Parse::EventParser::matches( $toks, qw( province = :province: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            # name, desc
            if ( EU2::Parse::EventParser::matches( $toks, qw( (name|desc) = .* ))) {
                $obj->{uc $toks->[0]->text} = $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            # style
            if ( EU2::Parse::EventParser::matches( $toks, qw( style = :int,0,5: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            # dates (TODO: Is enddate valid?)
            if ( EU2::Parse::EventParser::matches( $toks, qw( (date|deathdate) = { ))) {
                @tmpdate = (undef, undef, undef, lc $toks->[0]->text);  #yyyy, mm, dd, type
                $state = '<date>';
                splice (@$toks, 0, 3);
                next;
            }

#            # date, deathdate
#            if ( EU2::Parse::EventParser::matches( $toks, qw( (date|deathdate) = { day = :int,0,31: month = :month: year = :year: } ))) {
#                splice (@$toks, 0, 13);
#                next;
#            }
#
#            # date, deathdate #2 - different order
#            if ( EU2::Parse::EventParser::matches( $toks, qw( (date|deathdate) = { year = :year: month = :month: day = :int,0,31: } ))) {
#                splice (@$toks, 0, 13);
#                next;
#            }
#
#            # date, deathdate #2 -- just year
#            if ( EU2::Parse::EventParser::matches( $toks, qw( (date|deathdate) = { year = :year: } ))) {
#                splice (@$toks, 0, 7);
#                next;
#            }
#
            # offset
            if ( EU2::Parse::EventParser::matches( $toks, qw( offset = :int,0,inf: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            # trigger block
            if ( EU2::Parse::EventParser::matches( $toks, qw( trigger = { ))) {
                $state = '<trigger>';
                splice (@$toks, 0, 3);
                next;
            }

            # action block
            if ( EU2::Parse::EventParser::matches( $toks, qw( action_[abcd] = { ))) {
                $state = '<action>';
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( } ))) {
                $state = '<end>';
                splice (@$toks, 0, 1);
                next;
            }
            #....................
            do {
                EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                EU2::Parse::EventParser::balanceParen( $toks, '{', '}');
                $state = '<end>';
            }

        }

        elsif ($state eq '<date>') {
            if ( EU2::Parse::EventParser::matches( $toks, qw( } ))) {
                if (!defined $tmpdate[0]) {
                    $obj->Log("Missing year in date -- assuming 1419: ");
                    EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                    $tmpdate[0] = 1419;
                }
                $tmpdate[1] = 'january' if (!defined $tmpdate[1]);    #default month
                $tmpdate[2] = 1 if (!defined $tmpdate[2]);          #default day

                #my $date = EU2::Parse::EventParser::formatDate( @tmpdate[0,1,2] );
                #if ($tmpdate[3] eq 'startdate') {
                #    $obj->{STARTDATE} = $date;
                #} elsif ($tmpdate[3] =~ /^(deathdate|enddate)$/) {
                #    $obj->{DEATHDATE} = $date;
                #} else {
                #    $obj->Log("{Program error} Expected startdate, deathdate, or enddate, not '$tmpdate[3]'");
                #}

                $state = '<event>';
                splice (@$toks, 0, 1);
                next;

            }
            if ( EU2::Parse::EventParser::matches( $toks, qw( year = :year: ))) {
                if (defined $tmpdate[0]) {
                    $obj->Log("Redefinition of year, overwriting previous value: ");
                    EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                }
                $tmpdate[0] = 1 * $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( month = :month: ))) {
                if (defined $tmpdate[1]) {
                    $obj->Log("Redefinition of month, overwriting previous value: ");
                    EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                }
                $tmpdate[1] = $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( day = :int,0,31: ))) {
                if (defined $tmpdate[2]) {
                    $obj->Log("Redefinition of day, overwriting previous value: ");
                    EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                }
                $tmpdate[2] = 1 * $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }


            do {
                EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                EU2::Parse::EventParser::balanceParen( $toks, '{', '}');
                $state = '<event>';

            }
        }

        elsif ($state eq '<trigger>') {
            if ( EU2::Parse::EventParser::matches( $toks, qw( } ))) {
                #$state = '<event>';
                $state = ((pop @state_stack) || '<event>');
                splice (@$toks, 0, 1);
                next;
            }

            ## Note: Doesn't check for 2 or more triggers for AND/OR
            if ( EU2::Parse::EventParser::matches( $toks, qw( (and|or|not) = { ))) {
                push @state_stack, $state;
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( event = :int,0,4294967296: ))) {
                push @{$obj->{EVENT_REFS}}, 1*$toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( year = :year: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( (alliance|dynastic|vassal|war) = { country = :ctag: country = :ctag: } ))) {
                splice (@$toks, 0, 10);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( atwar = (yes|no) ))) {
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( provinceculture = { prov = :province: data = :culture: } ))) {
                splice (@$toks, 0, 10);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( provincereligion = { province = :province: data = :religion: } ))) {
                splice (@$toks, 0, 10);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( religion = :religion: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( (leader|monarch) = :int,0,200000: ))) {
                ## Add to the leader_refs or monarch_refs
                if ($toks->[0]->text eq 'leader') {
                    push @{$obj->{LEADER_REFS}}, 1*$toks->[2]->text;
                } elsif ($toks->[0]->text eq 'monarch') {
                    push @{$obj->{MONARCH_REFS}}, 1*$toks->[2]->text;
                }
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( (badboy|countrysize) = :int,0,inf: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( continent = :continent: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            ## style 1 of control/owned trigger: missing data means the country defined for the event
            if ( EU2::Parse::EventParser::matches( $toks, qw( (control|owned) = { province = :province: } ))) {
                splice (@$toks, 0, 7);
                next;
            }

            ## style 2 of control/owned trigger
            if ( EU2::Parse::EventParser::matches( $toks, qw( (control|owned) = { province = :province: data = :ctag,-1: } ))) {
                splice (@$toks, 0, 10);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( core = { province = :province: data = :ctag,-1: } ))) {
                #The core trigger is fixed in 1.06
                #EU2::Parse::EventParser::Log 0, sprintf("Note: The core trigger is reputed to not work (line %d char %d)", $toks->[0]->location);
                splice (@$toks, 0, 10);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( (cot|discovered) = :province: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( (exists|neighbour) = :ctag: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( relation = { country = :ctag: value = :int,-200,200: } ))) {
                splice (@$toks, 0, 10);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( domestic = { type = :diplo: value = :int,0,10: } ))) {
                splice (@$toks, 0, 10);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( (infra|trade) = :int,0,10: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( (land|naval) = :int,0,60: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( stability = :int,-15v,15v: ))) {
                splice (@$toks, 0, 3);
                next;
            }

            ## New for 1.06
            if ( EU2::Parse::EventParser::matches( $toks, qw( ai = (yes|no) ))) {
                splice (@$toks, 0, 3);
                next;
            }

            ## New for 1.06
            if ( EU2::Parse::EventParser::matches( $toks, qw( flag = .* ))) {
                splice (@$toks, 0, 3);
                next;
            }

            do {
                EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                EU2::Parse::EventParser::balanceParen( $toks, '{', '}');
                #$state = '<event>';
                $state = ((pop @state_stack) || '<event>');
            }
        }

        elsif ($state eq '<action>') {
            if ( EU2::Parse::EventParser::matches( $toks, qw( } ))) {
                $state = '<event>';
                splice (@$toks, 0, 1);
                next;
            }
            if ( EU2::Parse::EventParser::matches( $toks, qw( name = .* ))) {
                push @{$obj->{ACTIONNAME_REFS}}, $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( command = { ))) {
                $state = '<action_command>';
                splice (@$toks, 0, 3);
                next;
            }

            do {
                EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                EU2::Parse::EventParser::balanceParen( $toks, '{', '}');
                $state = '<event>';

            }
        }

        elsif ($state eq '<action_command>') {
            if ( EU2::Parse::EventParser::matches( $toks, qw( } ))) {
                $state = '<action>';
                splice (@$toks, 0, 1);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (alliance|breakdynastic|breakvassal|dynastic|independence|inherit|vassal|war) which = :ctag,-1: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = casusbelli which = :ctag,-1: value = :int,0,inf: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = relation which = :ctag,-1: value = :int,-400,400: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (conversion|heretic|religiousrevolt) which = :province,rand: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = provincereligion which = :province,rand: value = :religion: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = religion which = :religion: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (add_countryculture|remove_countryculture) which = :culture: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = provinceculture which = :province,rand: value = :culture,-1: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (conquistador|explorer) which = :province,rand: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (leader|monarch|sleepleader|sleepmonarch|wakeleader|wakemonarch) which = :int,0,200000: ))) {
                if ($toks->[2]->text =~ /leader$/i) {
                    push @{$obj->{LEADER_REFS}}, 1*$toks->[5]->text;
                } elsif ($toks->[2]->text =~ /monarch$/i) {
                    push @{$obj->{MONARCH_REFS}}, 1*$toks->[5]->text;
                }
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (adm|mil|dip) which = :int,-15,15: value = :int,0,inf: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = country which = :ctag: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = technology which = :tech: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (capital|addcore|removecore|cot) which = :province: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = secedeprovince which = :ctag: value = :province: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = province_revoltrisk which = :province: value = :int,-100,100: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = population which = :province,rand: value = :int,-inf,inf: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = provincetax which = :province,rand: value = :int,-10v,10v: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = manpower value = :int,-100v,100v: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = provincemanpower which = :province,rand: value = :int,-10v,10v: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = natives which = :province: value = :int,0,9: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = mine which = :province,rand: value = :int,-200v,200v: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = domestic which = :diplo: value = :int,-10,10: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            # badboy added in 1.07
            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (infra|trade|land|naval|treasury|cash|inflation|vp|badboy) value = :int,-inf,inf: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = loansize which = :int,1,inf: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = stability value = :int,-15v,15v: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (revolt|colonialrevolt) which = :province,rand: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = revoltrisk which = :province,rand: value = :int,-100,100: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            #revoltrisk #2 - missing months defaults to 12
            if ( EU2::Parse::EventParser::matches( $toks, qw( type = revoltrisk value = :int,-100,100: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = fortress which = :province,rand: value = :int,-6,6: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (gainbuilding|losebuilding) which = :province,rand: value = :building: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = gainmanufactory which = :province,rand: value = :manufactury: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = losemanufactory which = :province,rand: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (diplomats|colonists|merchants|missionaries) value = :int,-6,6: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (desertion|inf|cav|art|warships|galleys|transports|infantry|cavalry|artillery) which = :province,rand: value = :int,0,inf: ))) {
                splice (@$toks, 0, 9);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = flag which = :int,0,6: ))) {
                splice (@$toks, 0, 6);
                next;
            }

            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (trigger|sleepevent) which = :int,0,4294967296: ))) {
                push @{$obj->{EVENT_REFS}}, 1*$toks->[5]->text;
                splice (@$toks, 0, 6);
                next;
            }

            ## New in 1.06
            if ( EU2::Parse::EventParser::matches( $toks, qw( type = (setflag|clrflag|ai) which = .* ))) {
                splice (@$toks, 0, 6);
                next;
            }


            do {
                EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                EU2::Parse::EventParser::balanceParen( $toks, '{', '}');
                $state = '<action>';
            }
        }

        else {
                EU2::Parse::EventParser::warnBadState(ref $obj, $state, $toks);
                EU2::Parse::EventParser::balanceParen( $toks, '{', '}');
                $state = '<end>';

        }
    } #while

    ## Make sure all the required params are defined:
    #my @missing = ();
    #foreach my $field ( qw(ID NAME CATEGORY STARTDATE DEATHDATE RANK MOVEMENT FIRE SHOCK SIEGE) ) {
    #   (defined $obj->{$field}) or push @missing, lc($field);
    #}
    #if (@missing) {
    #   EU2::Parse::LeaderParser::Log(0, "Warning: Missing definition for " . join(",", @missing) . " in event at line $startline");
    #}

    1;

}

1;
