#!/usr/bin/perl

#---------------------------------------------------------------------
# Copyright (C) 2003 Chad Redman <chad_eu2@katica.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#---------------------------------------------------------------------


# File:     EU2/Parse/MonarchParser.pm
# Author:   Chad Redman <chad_eu2@katica.org>

# $Id: MonarchParser.pm,v 1.2 2003/07/22 04:45:04 chad Exp $

package EU2::Parse::MonarchParser;
use EU2::Parse::GenericParser;

@ISA= ('EU2::Parse::GenericParser');

sub new {
    my ($pkg) = @_;
    my $self = EU2::Parse::GenericParser::new();
    $self->{MONARCHS} = [];

    return bless $self, $pkg;
}


sub monarchs {  my $self = shift; return $self->{MONARCHS}; }

sub addMonarch {  my ($self, $id) = @_; push(@{$self->{MONARCHS}}, $id) }


sub matches      { return EU2::Parse::GenericParser::matches('EU2::Parse::MonarchParser', @_); }
sub warnBadState { return EU2::Parse::GenericParser::warnBadState(@_); }
sub balanceParen { return EU2::Parse::GenericParser::balanceParen(@_); }
sub formatDate { return EU2::Parse::GenericParser::formatDate(@_); }
#sub optCheckRange { return EU2::Parse::GenericParser::optCheckRange(@_); }
#sub _abortParseError { return EU2::Parse::GenericParser::_abortParseError(@_); }


sub parse {
    my $self = shift;
    $self->_tokenize;

    my $toks = $self->{symbols};

    while (@$toks) {
        my $monarch = Monarch->new;
        if ($monarch->parse($toks)) {
            $self->addMonarch( $monarch );
        } else {
            last;
            #Log(0, "DEBUG? Warning: Expected 'historicalleader = {' at line " . $toks->[0]->line . " char " . $toks->[0]->char);
            #warnBadState(ref $self, '-TOP-', $toks);
            #return _abortParseError($toks);
        }
    }
}

sub tokenMatches {
    ## Nothing special here

    -1;    # Indicates caller can continue its own attempts to match
}



#-------------------------------------
package Monarch;

sub new {
    my $pkg = shift;
    my $self = {
        'ID' => undef,
        'LINE' => undef,    # Line no. on which the block starts
        'NAME' => undef,
        'STARTDATE' => undef,
        'DEATHDATE' => undef,
        'DIP' => undef,
        'ADM' => undef,
        'MIL' => undef,
        #-- optional--
        'DORMANT' => undef,
        'REMARK' => undef,
    };

    return bless $self, $pkg;
}

sub Log {
    my $obj = shift;
    EU2::Parse::MonarchParser->Log(@_);
}

sub parse {
    my ($obj, $toks) = @_;
    my $state = '<start>';
    my $startline = $toks->[0]->line;
    $obj->{LINE} = $startline;
    my @tmpdate;  ## Store the YMD of dates while parsing state <date>

    while ($state ne '<end>') {
        if (scalar @$toks == 0) {  ## nothing left in the whole file
            $state = '<end>';
            next;
        }

        if ($state eq '<start>') {
            if ( EU2::Parse::MonarchParser::matches( $toks, qw(historicalmonarch = { ))) {
                $state = '<monarch>';
                splice (@$toks, 0, 3);
            } else {
                ## Scan remaining tokens for a positive match
                EU2::Parse::MonarchParser::warnBadState(ref $obj, $state, $toks);
                while (@$toks) {
                    if ( EU2::Parse::MonarchParser::matches( $toks, qw(historicalmonarch = { ))) {
                        last;
                    }
                    shift @$toks;
                }
                @$toks or $obj->Log("No monarch found before end of file");
                #$state = '<end>';
                return 0;
            }
        }

        elsif ($state eq '<monarch>') {
            if ( EU2::Parse::MonarchParser::matches( $toks, qw( id = { type = :int,6,6: id = :int,0,200000: } ))) {
                $obj->{ID} = 1 * $toks->[8]->text;   # convert to number
                splice (@$toks, 0, 10);
                next;
            }

            if ( EU2::Parse::MonarchParser::matches( $toks, qw( name = .* ))) {
                $obj->{NAME} = $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            # dates (TODO: Is enddate valid?)
            if ( EU2::Parse::MonarchParser::matches( $toks, qw( (startdate|deathdate|enddate) = { ))) {
                @tmpdate = (undef, undef, undef, lc $toks->[0]->text);  #yyyy, mm, dd, type
                $state = '<date>';
                splice (@$toks, 0, 3);
                next;
            }

            # other single-valued params
            if ( EU2::Parse::MonarchParser::matches( $toks, qw( (dip|adm|mil) = :int,0,9: ))) {
                $obj->{uc($toks->[0]->text)} = $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            # optional params
            if ( EU2::Parse::MonarchParser::matches( $toks, qw( dormant = (yes|no) ))) {
                $obj->{uc($toks->[0]->text)} = $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::MonarchParser::matches( $toks, qw( remark = .* ))) {
                $obj->{uc($toks->[0]->text)} = $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::MonarchParser::matches( $toks, qw( } ))) {
                $state = '<end>';
                splice (@$toks, 0, 1);
                next;
            }
            #....................
            do {
                EU2::Parse::MonarchParser::warnBadState(ref $obj, $state, $toks);
                EU2::Parse::MonarchParser::balanceParen( $toks, '{', '}');
                $state = '<end>';
            }

        }

        elsif ($state eq '<date>') {
            if ( EU2::Parse::MonarchParser::matches( $toks, qw( } ))) {
                if (!defined $tmpdate[0]) {
                    $obj->Log("Missing year in date -- assuming 1419: ");
                    EU2::Parse::MonarchParser::warnBadState(ref $obj, $state, $toks);
                    $tmpdate[0] = 1419;
                }
                $tmpdate[1] = 'january' if (!defined $tmpdate[1]);    #default month
                $tmpdate[2] = 1 if (!defined $tmpdate[2]);          #default day

                my $date = EU2::Parse::MonarchParser::formatDate( @tmpdate[0,1,2] );
                if ($tmpdate[3] eq 'startdate') {
                    $obj->{STARTDATE} = $date;
                } elsif ($tmpdate[3] =~ /^(deathdate|enddate)$/) {
                    $obj->{DEATHDATE} = $date;
                } else {
                    $obj->Log("{Program error} Expected startdate, deathdate, or enddate, not '$tmpdate[3]'");
                }

                $state = '<monarch>';
                splice (@$toks, 0, 1);
                next;

            }
            if ( EU2::Parse::MonarchParser::matches( $toks, qw( year = :year: ))) {
                if (defined $tmpdate[0]) {
                    $obj->Log("Redefinition of year, overwriting previous value: ");
                    EU2::Parse::MonarchParser::warnBadState(ref $obj, $state, $toks);
                }
                $tmpdate[0] = 1 * $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::MonarchParser::matches( $toks, qw( month = :month: ))) {
                if (defined $tmpdate[1]) {
                    $obj->Log("Redefinition of month, overwriting previous value: ");
                    EU2::Parse::MonarchParser::warnBadState(ref $obj, $state, $toks);
                }
                $tmpdate[1] = $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }

            if ( EU2::Parse::MonarchParser::matches( $toks, qw( day = :int,0,31: ))) {
                if (defined $tmpdate[2]) {
                    $obj->Log("Redefinition of day, overwriting previous value: ");
                    EU2::Parse::MonarchParser::warnBadState(ref $obj, $state, $toks);
                }
                $tmpdate[2] = 1 * $toks->[2]->text;
                splice (@$toks, 0, 3);
                next;
            }


            do {
                EU2::Parse::MonarchParser::warnBadState(ref $obj, $state, $toks);
                EU2::Parse::MonarchParser::balanceParen( $toks, '{', '}');
                $state = '<monarch>';

            }
        }

        else {
                EU2::Parse::MonarchParser::warnBadState(ref $obj, $state, $toks);
                EU2::Parse::MonarchParser::balanceParen( $toks, '{', '}');
                $state = '<end>';

        }
    } #while


    ## Make sure all the required params are defined:
    my @missing = ();
    foreach my $field ( qw(ID NAME  STARTDATE DEATHDATE DIP ADM MIL) ) {
        (defined $obj->{$field}) or push @missing, lc($field);
    }
    if (@missing) {
        $obj->Log("Warning: Missing definition for " . join(",", @missing) . " in event at line $startline");
    }

    1;

}

1;
