@REM ----------------------------------------------------------------------
@REM Batch script for analyzing the EEP mod. This assumes that the EEP moddir is
@REM directly under the EU2 base directory, and that the EULint directory is in
@REM a subdirectory of EU2 called EU2Lint. The output log is set up to be
@REM C:\Temp\output.txt, while the tab-delimited data files will be saved to
@REM the C:\Temp\lint_data directory. Any of these values can be changed to suit
@REM a different environment
@REM ----------------------------------------------------------------------

@REM -----------------------------------------------------------------------
@SET EU2=C:\Program Files\Strategy First\Europa Universalis 2
@SET MODDIR=%EU2%\EEP
@SET LINT=%EU2%\EU2Lint
@SET OUTPUT=C:\Temp\output.txt
@REM -----------------------------------------------------------------------


CHDIR C:\Temp


REM ---------------------------------------------------------
REM Parse leader data and save to lint_data\leaders.txt ...
REM ---------------------------------------------------------
perl -I"%LINT%" "%LINT%\lint.pl" -leaders "%MODDIR%\db\leaders" -countrycsv "%MODDIR%\db\country.csv" > %OUTPUT%


REM ---------------------------------------------------------
REM Parse monarch data and save to lint_data\monarchs.txt.
REM Use the existing leaders.txt as a reference for leader IDs
REM ---------------------------------------------------------
perl -I"%LINT%" "%LINT%\lint.pl" -importleaders -monarchs "%MODDIR%\db\monarchs" -countrycsv "%MODDIR%\db\country.csv" >> %OUTPUT%


REM --------------------------------------------------------------
REM Parse event data and save to lint_data\events.txt.
REM Use the existing leaders.txt and monarchs.txt as references
REM --------------------------------------------------------------
perl -I"%LINT%" "%LINT%\lint.pl" -importleaders -importmonarchs -countrycsv "%MODDIR%\db\country.csv" -textcsv "%MODDIR%\Config\text.csv" -eventlist "%MODDIR%\db\events.txt" -events "%MODDIR%\db\Events" >> %OUTPUT%

notepad.exe "%OUTPUT%"
pause

exit

@REM ----END----

@REM #Equivalent for Unix
@REM lint=/home/chad/dev/eu2/EU2Lint
@REM eep=/home/chad/dev/eu2/mods/eep141
@REM
@REM echo $(date) "Started $0"
@REM echo "--------------------------------------"
@REM perl -I"$lint" $lint/lint.pl -leaders $eep/db/leaders -monarchs $eep/db/monarchs -countrycsv \
@REM   $eep/db/country.csv -textcsv $eep/Config/text.csv -eventlist $eep/db/events.txt -events $eep/db/Events
@REM echo "--------------------------------------"
@REM echo $(date) "Finished $0"
